# -*- coding: utf-8 -*-
"""Account class module for HekrAPI"""

from typing import Union
from aiohttp import ClientSession

from .device import Device
from .const import DEFAULT_APPLICATION_ID
from .exceptions import AccountUnauthenticatedException, AccountDevicesUpdateFailedException


class Account:
    """Account class for HekrAPI

    Raises:
        ValueError: Account constructor is not provided possible ways to authenticate
        AccountUnauthenticatedException: Account did not authenticate prior to calling method

    Attributes:
        __username (str): Account username
        __password (str, optional): Account password (not required with token)
        __token (str, optional): Authentication 'Bearer' token (not required with password)
        __devices (dict): Dictionary with 'Device' objects belonging to account
        application_id (str): Application ID (has default value)
    """

    BASE_URL = 'https://user-openapi.hekreu.me'

    def __init__(self, username: str, password: Union[str, type(None)]=None,
                 token: Union[str, type(None)]=None,
                 application_id: str = DEFAULT_APPLICATION_ID):
        # @TODO: refactor after enabling authentication support
        if not token:  # and not password:
            raise ValueError(
                "at least one of arguments (password, token) should be set")

        self.application_id = application_id

        self.__username = username
        self.__password = password
        self.__token = token

        self.__devices = {}

    @property
    def devices(self):
        """Device dictionary accessor

        Returns:
            dict -- dictionary of devices (device_id => Device object)
        """
        return self.__devices

    def authenticate(self):
        """Authenticate account with Hekr"""

    async def update_devices(self):
        """Update devices for account

        Raises:
            AccountUnauthenticatedException: [description]

        Returns:
            int -- New devices found (that do not exist within the __devices attribute)
        """

        if not self.__token:
            raise AccountUnauthenticatedException()

        base_url_devices = self.BASE_URL + '/devices?size={}&page={}'

        request_devices = 20
        current_page = 0

        new_devices = {}

        async with ClientSession() as session:
            more_devices = True
            while more_devices:
                request_url = base_url_devices.format(
                    request_devices,
                    current_page
                )

                async with session.get(
                        request_url,
                        headers={'Authorization': 'Bearer ' + self.__token}
                ) as response:
                    response_json = await response.json()

                    if response.status != 200:
                        session.close()
                        raise AccountDevicesUpdateFailedException(self, response)

                    more_devices = (len(response_json) == request_devices)

                    for device_attributes in response_json:
                        device_id = device_attributes['devTid']

                        if device_id in self.__devices:
                            self.__devices[device_id].set_control_key(
                                device_attributes['ctrlKey'])
                        else:
                            new_devices[device_id] = Device(
                                device_id=device_id,
                                control_key=device_attributes['ctrlKey'],
                                host=device_attributes['lanIp'],
                                application_id=self.application_id
                            )

            if new_devices:
                self.__devices.update(new_devices)

        return len(new_devices)



# -*- coding: utf-8 -*-
"""Device class module for Hekr API"""
from typing import Union

from enum import IntEnum
from socket import (
    socket, AF_INET, SOCK_DGRAM,
    timeout as SocketTimeout
)
from json import (dumps, loads)

from .command import Command
from .exceptions import (
    DeviceProtocolNotSetException,
    HeartbeatFailedException,
    LocalAuthenticationFailedException,
    CloudAuthenticationFailedException,
    CommandFailedException,
    DeviceConnectionMissingException,
    AuthenticationFailedException
)
from .const import (
    DEFAULT_APPLICATION_ID,
    ACTION_AUTHENTICATE,
    DEFAULT_DEVICE_PORT
)


def device_id_from_mac_address(mac_address: Union[str, bytearray]) -> str:
    """Convert mac address to device ID

    Arguments:
        mac_address {Union[str, bytearray]} -- MAC-address string (dashed format)

    Returns:
        str -- Device ID
    """
    if isinstance(mac_address, bytearray):
        mac_address = mac_address.hex()

    for delimiter in [':', '-', ' ']:
        mac_address = mac_address.replace(delimiter, '')

    mac_address = mac_address.upper()
    return 'ESP_2M_' + mac_address


class DeviceConnectionType(IntEnum):
    """Connection types for devices"""
    NONE = 0
    LOCAL = 1
    CLOUD = 2


class Device:
    """Device class for Hekr API"""

    def __init__(self,
                 device_id: str,
                 control_key: str,
                 account: Union['Account', type(None)]=None,
                 host: Union[str, type(None)]=None,
                 port: int = DEFAULT_DEVICE_PORT,
                 application_id: str = DEFAULT_APPLICATION_ID,
                 protocol: Union['Protocol', type(None)]=None
                 ):
        self.device_id = device_id
        self.local_address = (host, port)
        self.account = account
        self.protocol = protocol
        self.application_id = application_id
        self.local_socket = None

        self.local_message_id = 0

        self.__local_authenticated = False
        self.__control_key = control_key

    @property
    def available_connection_type(self) -> DeviceConnectionType:
        """Retrieve best available connection type for communication

        Returns:
            DeviceConnectionType -- Available connection type
        """
        # @TODO: check for more stuff to determine connection type
        if self.local_address and self.__control_key:
            return DeviceConnectionType.LOCAL
        if self.account:
            return DeviceConnectionType.CLOUD

        return DeviceConnectionType.NONE

    def set_control_key(self, control_key: str):
        """Sets device control key

        Arguments:
            control_key {str} -- Control key
        """
        self.__control_key = control_key

    def open_socket_local(self):
        """Opens local socket to device"""
        sock = socket(AF_INET, SOCK_DGRAM)
        sock.connect(self.local_address)
        sock.settimeout(5)
        # sock.setblocking(0)
        self.local_socket = sock

    def make_request(self, action: str, params: dict = None,
                     message_id: int = None,
                     frame_number: int = None,
                     connection_type: DeviceConnectionType = None):
        """Make device request

        Arguments:
            action {str} -- Action string

        Keyword Arguments:
            params {dict} -- Request parameters (default: {None})
            message_id {int} -- Message ID (default: {None})
            frame_number {int} -- Frame number (default: {None})
            connection_type {DeviceConnectionType} -- Connection type (default: {None})

        Returns:
            dict -- Request response
        """
        connection_type = connection_type or self.available_connection_type

        if connection_type == DeviceConnectionType.LOCAL:
            # @TODO: message IDs should be carried across all instances by definition
            if not self.__local_authenticated and action != ACTION_AUTHENTICATE:
                self.authenticate(connection_type=DeviceConnectionType.LOCAL)

            if self.local_socket is None:
                self.open_socket_local()

            if message_id is None:
                self.local_message_id += 1
                message_id = self.local_message_id

            if frame_number is None:
                frame_number = 1

            request_dict = {
                "msgId": message_id,
                "action": action,
                "params": {
                    "ctrlKey": self.__control_key,
                    "devTid": self.device_id
                }
            }
            if params:
                request_dict["params"].update(params)

            self.local_socket.send(dumps(request_dict).encode('utf-8'))

            return self.read_response(connection_type=connection_type)

        if connection_type == DeviceConnectionType.CLOUD:
            # @TODO: not implemented
            return False

        raise DeviceConnectionMissingException(self)

    def read_response(self, length: int = 256, connection_type: DeviceConnectionType = None):
        """Read device response

        Keyword Arguments:
            length {int} -- Message length to expect (default: {256})
            connection_type {DeviceConnectionType} -- Connection type to use (default: {None})

        Returns:
            dict -- JSON-decoded dictionary of values
        """
        connection_type = connection_type or self.available_connection_type

        if connection_type == DeviceConnectionType.LOCAL:
            received_data = self.local_socket.recv(length)

            # @TODO: exceptions for nothingness
            return loads(received_data)

        if connection_type == DeviceConnectionType.CLOUD:
            # @TODO: not implemented
            pass

        return DeviceConnectionMissingException(self)

    def heartbeat(self, connection_type: DeviceConnectionType = None):
        """Send heartbeat message

        Keyword Arguments:
            connection_type {DeviceConnectionType} -- Connection type to use (default: {None})

        Raises:
            HeartbeatFailedException: Heartbeat message sending failed
        """
        response = self.make_request(
            action='heartbeat',
            params=False,
            connection_type=connection_type)

        if response.get("code", None) != 200:
            raise HeartbeatFailedException(self, response)

    def command(self,
                command: Union[int, str, Command],
                data: dict = None,
                return_decoded=True):
        """Execute device command and return response

        Arguments:
            command {Union[int, str, Command]} -- Command ID/name/object to use

        Keyword Arguments:
            data {dict} -- Data values for datagram (default: {None})
            return_decoded {bool} -- Extract and decode datagram from response (default: {True})

        Raises:
            DeviceProtocolNotSetException: Device does not have a protocol set
            CommandFailedException: Command failed to execute due to device code error
            CommandFailedException: Command failed to execute due to missing datagram
            CommandFailedException: Command failed to execute due to socket timeout

        Returns:
            dict -- decoded datagram values or full response dictionary
        """
        if not isinstance(command, Command):
            if self.protocol:
                command = self.protocol.get_command(command)
            else:
                raise DeviceProtocolNotSetException(self)

        raw = self.protocol.encode(
            data=data,
            command=command,
            frame_number=1
        )

        request_dict = {
            "appTid": self.application_id,
            "data": {
                "raw": raw
            }
        }
        response = self.make_request(
            action='appSend',
            params=request_dict
        )

        if response.get("code", None) != 200:
            raise CommandFailedException(command, self, response)

        try:
            response = self.read_response(512)

            if return_decoded:
                datagram = (response
                            .get('params', {})
                            .get('data', {})
                            .get('raw', None)
                            )

                if not datagram:
                    raise CommandFailedException(
                        command, self, response, reason='Datagram not found')

                return self.protocol.decode(datagram)

            return response

        except SocketTimeout:
            raise CommandFailedException(
                command, self, response, reason='Socket timeout')

    def authenticate(self, connection_type: DeviceConnectionType = None):
        """Authenticate with the device

        Keyword Arguments:
            connection_type {DeviceConnectionType} -- Connection type to use (default: {None})

        Raises:
            LocalAuthenticationFailedException: Local authentication failed
            AuthenticationFailedException: Authentication failed
        """
        connection_type = connection_type or self.available_connection_type

        if connection_type == DeviceConnectionType.LOCAL:
            response = self.make_request(
                action=ACTION_AUTHENTICATE,
                connection_type=connection_type)

            if response.get("code", None) != 200:
                raise LocalAuthenticationFailedException(self, response)

            self.__local_authenticated = True
            return True

        elif connection_type == DeviceConnectionType.CLOUD:
            raise CloudAuthenticationFailedException(self)

        # @TODO: research into authentication
        raise AuthenticationFailedException(self)



# -*- coding: utf-8 -*-
"""Datagram encoding/decoding module for HekrAPI"""

from typing import Union, Tuple

from .exceptions import (
    InvalidMessagePrefixException,
    InvalidMessageLengthException,
    InvalidMessageChecksumException,
    InvalidMessageFrameTypeException,
    InvalidDataMissingKeyException
)
from .const import FRAME_START_IDENTIFICATION

def decode(protocol: 'Protocol', raw: Union[bytearray, str],
           use_variable_names=False,
           filter_values=True
           ) -> Tuple['Command', dict, int]:
    """Decode raw Hekr datagram

    Arguments:
        protocol {[type]} -- [description]
        raw {Union[bytearray, str]} -- [description]

    Keyword Arguments:
        use_variable_names {bool} -- Use variable names when decoding (default: {False})
        filter_values {bool} -- Apply multiplication and typecasting to values (default: {True})

    Raises:
        TypeError: Invalid type provided as function argument
        InvalidMessagePrefixException: Datagram prefix does not match expected
        InvalidMessageLengthException: Datagram length does not match expected
        InvalidMessageChecksumException: Datagram checksum does not match expected
        InvalidMessageFrameTypeException: Datagram frame type does not match expected

    Returns:
        dict -- Dictionary of decoded values
    """
    if isinstance(raw, str):
        decoded = bytearray.fromhex(raw)
    elif isinstance(raw, bytearray):
        decoded = raw
    else:
        raise TypeError(type(raw))

    if decoded[0] != FRAME_START_IDENTIFICATION:
        raise InvalidMessagePrefixException(raw)

    frame_length = decoded[1]
    if frame_length != len(decoded):
        raise InvalidMessageLengthException(raw)

    checksum = decoded[-1]
    current_checksum = sum(decoded[:-1]) % 0x100
    if checksum != current_checksum:
        raise InvalidMessageChecksumException(raw)

    frame_type = decoded[2]
    command_id = decoded[4]
    command = protocol.get_command_by_id(command_id)
    if frame_type != command.frame_type:
        raise InvalidMessageFrameTypeException(raw)

    frame_number = decoded[3]

    current_pos = 5
    data = {}
    for argument in command.arguments:
        next_pos = current_pos + argument.byte_length
        value_output = int.from_bytes(
            decoded[current_pos:next_pos], byteorder='big', signed=False)

        if filter_values:
            if argument.multiplier is not None:
                value_output *= argument.multiplier
                if argument.decimals is not None:
                    value_output = round(value_output, argument.decimals)

            value_output = argument.type_output(value_output)

        data[argument.variable if use_variable_names else argument.name] = value_output
        current_pos = next_pos

    return (command, data, frame_number)

def encode(command: 'Command', data: dict = None, frame_number: int = 1,
           use_variable_names=False, filter_values=True) -> str:
    """Encode raw Hekr datagram

    Arguments:
        command {[type]} -- Command to use arguments from

    Keyword Arguments:
        data {dict} -- Dictionary of decoded values (default: {None})
        frame_number {int} -- Datagram frame number (default: {1})
        use_variable_names {bool} -- Use variable names when encoding (default: {False})
        filter_values {bool} -- Apply multiplication and typecasting (default: {True})

    Raises:
        InvalidDataMissingKeyException: [description]

    Returns:
        str -- [description]
    """
    if data is None:
        data = {}

    raw = bytearray()
    raw.append(FRAME_START_IDENTIFICATION)
    raw.append(command.frame_type)
    raw.append(frame_number)
    raw.append(command.command_id)

    for argument in command.arguments:
        key = argument.variable if use_variable_names else argument.name
        value_input = data.get(key, None)

        if value_input is None:
            raise InvalidDataMissingKeyException(key)

        if filter_values:
            if argument.multiplier:
                value_input /= argument.multiplier

            value_input = round(value_input)

        # @TODO: make better performing solution
        result = value_input.to_bytes(
            argument.byte_length,
            byteorder='big',
            signed=False)
        raw.extend(result)

    raw.insert(1, len(raw) + 2)
    raw.append(sum(raw) % 0x100)

    return raw.hex().upper()



# -*- coding: utf-8 -*-
"""Command class module for HekrAPI"""

from enum import IntEnum

from .argument import Argument

class FrameType(IntEnum):
    """Datagram frame types (per Hekr documentation)"""
    RECEIVE = 0x01
    SEND = 0x02
    DEVICE = 0xFE
    ERROR = 0xFF

class Command:
    """Command class for HekrAPI"""

    def __init__(self,
                 command_id: int,
                 name: str,
                 frame_type: FrameType,
                 *args: Argument
                 ):
        """Command class constructor

        Arguments:
            command_id {int} -- Command identifier for datagrams
            name {str} -- Command name for calls
            frame_type {FrameType} -- Command frame type (per Hekr docs)
        """
        self.command_id = command_id
        self.name = name
        self.frame_type = frame_type
        self.arguments = list(args)

    def __repr__(self):
        """Friendly command string conversion

        Returns:
            str -- Friendly debug representation
        """
        return '<{}("{}", {}, {})>'.format(
            self.__class__.__name__,
            self.name,
            self.command_id,
            self.frame_type.name
        )

    def __str__(self):
        """Return name of the command

        Returns:
            str -- Command name
        """
        return self.name

    def __int__(self):
        """Return command identifier

        Returns:
            int -- Command identifier
        """
        return self.command_id

    def print_definition(self, prefix=''):
        """Prints command definition in YAML format

        Keyword Arguments:
            prefix {str} -- What to prefix every output line with (default: {''})
        """
        print(prefix + '{}:'.format(self.name))
        new_prefix = prefix + '  '

        for attr in ['command_id', 'frame_type']:
            value = self.__getattribute__(attr)

            if value is not None:
                if isinstance(value, type):
                    value = value.__name__

                print(new_prefix + '{}: {}'.format(attr, value))
        print(new_prefix + 'arguments:')
        for argument in self.arguments:
            argument.print_definition(new_prefix + '  ')


DEFAULT_QUERY_COMMAND = Command(0, "queryDev", FrameType.SEND)



# -*- coding: utf-8 -*-
# pylint: disable=too-many-locals
"""Global constants for HekrAPI module"""

#: Default application ID used within requests via cloud and local transports
DEFAULT_APPLICATION_ID = 'hekr_python_connector'

#: Authentication action name for Account and Device authentication requests
ACTION_AUTHENTICATE = 'appDevAuth'

#: First character of every presumably valid raw datagram
FRAME_START_IDENTIFICATION = 0x48

#: Default port to send UDP datagrams to
DEFAULT_DEVICE_PORT = 10000



# -*- coding: utf-8 -*-
# pylint: disable=too-many-instance-attributes,too-many-arguments
"""Argument class module for HekrAPI"""

from typing import Callable, Any, Union, Tuple
from enum import IntEnum

ArgumentType = Callable[[Any], Any]
class Argument:
    """Argument class for HekrAPI

    Returns:
        Argument -- Argument definition
    """

    def __init__(self, name: str,
                 value_type: Union[Tuple[ArgumentType, ArgumentType], ArgumentType],
                 byte_length: int,
                 variable: Union[str, type(None)]=None,
                 multiplier: Union[int, float, type(None)]=None,
                 decimals: Union[int, type(None)]=None,
                 value_min: Union[int, float, type(None)]=None,
                 value_max: Union[int, float, type(None)]=None,
                 ):
        """Argument constructor

        Arguments:
            name {str} -- Argument name
            value_type {Union[Tuple[ArgumentType, ArgumentType], ArgumentType]} -- Value type
            byte_length {int} -- Length of input value in bytes

        Keyword Arguments:
            variable {Union[str, type} -- Variable name
                                          (in absence, name argument is used)
                                          (default: {None})
            multiplier {Union[int, float, type} -- What to multiply input/divide output values by
                                                   (default: {None})
            decimals {Union[int, type} -- Decimals to round input value to
                                          (in absence, extracted from multiply argument)
                                          (default: {None})
            value_min {Union[int, float, type} -- [description] (default: {None})
            value_max {Union[int, float, type} -- [description] (default: {None})
        """
        self.name = name
        self.variable = variable or name
        self.byte_length = byte_length

        if isinstance(value_type, IntEnum):
            self.type_input = int
            self.type_output = value_type

        elif isinstance(value_type, tuple):
            self.type_input = value_type[0]
            self.type_output = value_type[1]

        else:
            self.type_input = value_type
            self.type_output = value_type

        self.min = value_min
        self.max = value_max
        self.multiplier = multiplier

        # @TODO: check if numeric
        if multiplier is not None and decimals is None:
            self.decimals = str(self.multiplier)[::-1].find('.')
        else:
            self.decimals = decimals

    def __repr__(self):
        """Overloaded argument definiton

        Returns:
            str -- Human-friendly argument definition
        """
        return '<{}({})>'.format(self.__class__.__name__, self.name)

    def print_definition(self, prefix=''):
        """Print argument definition in YAML format

        Keyword Arguments:
            prefix {str} -- What to prefix every line with (default: {''})
        """
        print(prefix + self.name + ':')
        new_prefix = prefix + '  '

        for attr in ['variable', 'type_input', 'type_output',
                     'multiplier', 'decimals', 'min', 'max']:
            argument_attribute_value = self.__getattribute__(attr)

            if argument_attribute_value is not None:
                if isinstance(argument_attribute_value, type):
                    argument_attribute_value = argument_attribute_value.__name__

                print(new_prefix + '{}: {}'.format(attr, argument_attribute_value))



from distutils.core import setup
setup(
  name = 'hekrapi',         # How you named your package folder (MyLib)
  packages = ['hekrapi'],   # Chose the same as "name"
  version = '0.0.1',      # Start with a small number and increase it with every change you make
  license='MIT',        # Chose a license from here: https://help.github.com/articles/licensing-a-repository
  description = 'Python Hekr IoT API bindings',
  author = 'Alexander Ryazanov',
  author_email = 'alryaz@xavux.com',
  url = 'https://github.com/alryaz/hekrapi-python',
  download_url = 'https://github.com/alryaz/hekrapi-python/archive/0.0.1.tar.gz',
  keywords = ['Hekr', 'API', 'Wisen', 'Smart Devices', 'IoT'],   # Keywords that define your package best
  install_requires=[
          'aiohttp',
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
  ],
)


# -*- coding: utf-8 -*-
# pylint: disable=too-many-arguments
"""Protocol class module for Hekr API"""
from typing import Union
from .command import Command, FrameType, DEFAULT_QUERY_COMMAND
from .exceptions import CommandNotFoundException
from .datagram import encode, decode


class Protocol:
    """Protocol class definition for Hekr API"""

    def __init__(self, *args):
        """Protocol class constructor

        Arguments:
            *args {Command} -- Commands included in the protocol
        """
        self.__commands = list(args)
        # @TODO: make this more lean or make a generic platform definition
        self.__commands.append(DEFAULT_QUERY_COMMAND)

    def __getitem__(self, key: Union[int, str]):
        """Retrieves command with a square bracket accessor

        Arguments:
            key {Union[int, str]} -- Command ID/name

        Returns:
            Command -- Found command object
        """
        return self.get_command(key)

    def print_definition(self, sort=False, prefix: str = ''):
        """Prints protocol definition in YAML format

        Keyword Arguments:
            sort {bool} -- Sort commands by frame types (default: {False})
            prefix {str} -- What to prefix every line with (default: {''})
        """
        print(prefix + 'protocol:')
        new_prefix = prefix + '  '

        if sort:
            first_command = True
            for frame_type in FrameType:
                command_printed = False

                for command in self.__commands:
                    if command.frame_type == frame_type:
                        if not command_printed:
                            command_printed = True

                        if first_command:
                            first_command = False

                        else:
                            print()

                        command.print_definition(new_prefix)
        else:
            for command in self.__commands:
                command.print_definition(new_prefix)

    def decode(self, raw: Union[bytearray, str],
               use_variable_names=False, filter_values=True) -> dict:
        """Protocol-oriented datagram decoding"""
        return decode(
            raw=raw,
            protocol=self,
            use_variable_names=use_variable_names,
            filter_values=filter_values
        )

    def encode(self,
               command: Union[int,
                              str,
                              Command],
               data: dict,
               frame_number: int,
               use_variable_names=False,
               filter_values=True) -> str:
        """Protocol-oriented datagram encoding"""
        return encode(
            command=self.get_command(command),
            data=data,
            frame_number=frame_number,
            use_variable_names=use_variable_names,
            filter_values=filter_values
        )

    def get_command_by_id(self, command_id: int) -> Command:
        """Get command by its ID

        Arguments:
            command_id {int} -- Command ID to search for

        Raises:
            CommandNotFoundException: Command not found by given ID

        Returns:
            Command -- Command object
        """
        for command in self.__commands:
            if command.command_id == command_id:
                return command

        raise CommandNotFoundException(command_id)

    def get_command_by_name(self, name: str) -> Command:
        """Get command by its name

        Arguments:
            name {str} -- Command name to search for

        Raises:
            CommandNotFoundException: Command not found by given name

        Returns:
            Command -- Command object
        """
        for command in self.__commands:
            if command.name == name:
                return command

        raise CommandNotFoundException(name)

    def get_command(self, command: Union[int, str, Command]) -> Command:
        """Get command by its ID/name (or cycle-return object)

        Arguments:
            command {Union[int, str, Command]} -- Command ID/name/object

        Raises:
            TypeError: Bad argument type provided

        Returns:
            Command -- Command object
        """
        if isinstance(command, Command):
            return command
        if isinstance(command, int):
            return self.get_command_by_id(command)
        if isinstance(command, str):
            return self.get_command_by_name(command)

        raise TypeError(
            "Argument 'command' (type %s) does not evaluate to any supported command type" %
            command)



# -*- coding: utf-8 -*-
"""Exception classes for Hekr API"""

class HekrAPIException(Exception):
    """Base exception class for all HekrAPI exceptions"""
    default_reason = 'HekrAPI Exception occured'

    def __init__(self, *args, **kwargs):
        self.arguments = list(args)
        arguments_text = ', '.join([str(a) for a in self.arguments])
        kwargs['reason'] = kwargs.get('reason') or self.default_reason
        super().__init__(
            kwargs['reason'].format(
                **kwargs) + ' (' + arguments_text + ')')


class InvalidMessageException(HekrAPIException):
    """Base exception class for other message-related exceptions"""
    default_reason = 'Invalid Hekr message prvided'


class InvalidMessagePrefixException(InvalidMessageException):
    """Raised when the first character within the datagram does not equate 0x48"""
    default_reason = 'Message prefix is not standard'


class InvalidMessageLengthException(InvalidMessageException):
    """Raised when datagram data length does not equate one provided by the datagram"""
    default_reason = 'Message length does not match expected'


class InvalidMessageChecksumException(InvalidMessageException):
    """Raised when the calculated checksum does not equate one provided by the datagram"""
    default_reason = 'Message checksum does not match expected'


class InvalidMessageFrameTypeException(InvalidMessageException):
    """Raised when deduced command object frame type does not equate one provided by the datagram"""
    default_reason = 'Frame type does not match expected'


class InvalidDataException(HekrAPIException):
    """Base exception class for other data-related exceptions"""
    default_reason = 'Invalid Hekr data dictionary provided'


class InvalidDataMissingKeyException(InvalidDataException):
    """Raised when a specific data key is missing from the provided data dictionary"""
    default_reason = 'Key "{data_key}" not found in provided data dictionary'


class CommandNotFoundException(HekrAPIException):
    """Raised when a command was not found by value"""
    default_reason = 'Could not find command'


class CommandFailedException(HekrAPIException):
    """Raised when command execution response contains an error"""
    default_reason = "Could not execute command '{command}' on '{device.id}', response: {response}"


class HeartbeatFailedException(HekrAPIException):
    """Raised when heartbeat command failed to execute"""
    default_reason = "Could not execute heartbeat on '{device.id}', response: {response}"


class AuthenticationFailedException(HekrAPIException):
    """Base exception class for other authentication-related exceptions"""
    default_reason = "Authentication failed"


class LocalAuthenticationFailedException(AuthenticationFailedException):
    """Raised when local authentication fails"""
    default_reason = "Local authentication failed"


class CloudAuthenticationFailedException(AuthenticationFailedException):
    """Raised when cloud authentication fails"""
    default_reason = "Cloud authentication failed"


class AccountUnauthenticatedException(HekrAPIException):
    """Raised when account is not authenticated during method call that requires authentication"""
    default_reason = 'Account unauthenticated'


class DeviceProtocolNotSetException(HekrAPIException):
    """Raised when device does not have a protocol set after creation"""
    default_reason = "Device does not have a protocol set"


class DeviceConnectionMissingException(HekrAPIException):
    """Raised when device call cannot be made due to missing connection parameters"""
    default_reason = 'Cannot make requests to device until connection parameters are provided'

class AccountDevicesUpdateFailedException(HekrAPIException):
    """Raised when devices update function encounters a status code other than OK"""
    default_reason = 'Update server returned an invalid response code'